#extension GL_ARB_gpu_shader5 : enable
// data1

flat varying vec3 ambientUp;
flat varying vec3 ambientDown;
flat varying vec3 ambientLeft;
flat varying vec3 ambientRight;
flat varying vec3 ambientB;
flat varying vec3 ambientF;
flat varying float ambientA;
varying vec2 texcoord;

#include "/lib/light.glsl"

#ifdef SSAO
    #include "/lib/ssao.glsl"
#endif

vec3 CookTorranceBRDF(vec3 normal, vec3 viewDir, vec3 lightDir, float perceptualSmoothness, vec3 F0)
{
    normal = normalize(normal);
    viewDir = normalize(viewDir);
    lightDir = normalize(lightDir);

    vec3 halfVec = normalize(lightDir + viewDir);

    float NoL = max(dot(normal, lightDir), 0.0);
    float NoV = max(dot(normal, viewDir), 0.0);
    float NoH = max(dot(normal, halfVec), 0.0);
    float VoH = max(dot(viewDir, halfVec), 0.0);

    // Handle cases where NoV is very small
    if (NoV <= 0.0)
    {
        return vec3(0.0); // No reflection if the view direction is below the surface
    }

    // Convert perceptual smoothness to roughness
    float roughness = clamp(pow(1.0 - perceptualSmoothness, 2.0), 0.0001, 0.9999);

    // Distribution function (GGX)
    // float alpha = roughness * roughness;
    float alpha = roughness;
    float alpha2 = alpha * alpha;
    float denom = NoH * NoH * (alpha2 - 1.0) + 1.0;
    float D = alpha2 / (PI * denom * denom);

    // Fresnel function (Schlick's approximation)
    vec3 F = F0 + (1.0 - F0) * pow(1.0 - VoH, 5.0);

    // Geometry function (Smith's method with GGX)
    float k = alpha / 2.0;
    float G1 = NoL / (NoL * (1.0 - k) + k);
    float G2 = NoV / (NoV * (1.0 - k) + k);
    float G = G1 * G2;

    // Cook-Torrance BRDF
    vec3 specular = (D * F * G) / (4.0 * NoL * NoV + 0.0001);

    return specular;
}
vec3 drawSun(float cosY, float sunInt, vec3 nsunlight, vec3 inColor)
{
    return inColor +
           nsunlight / 0.001 * pow(smoothstep(cos(0.0093084168595 * 3.2), cos(0.0093084168595 * 1.8), cosY), 3.) * 0.62;
}

vec2 tapLocation(int sampleNumber, float spinAngle, int nb, float nbRot)
{
    float alpha = float(sampleNumber + spinAngle / 6.28) / nb;
    float angle = alpha * (nbRot * 6.28) + spinAngle;

    float ssR = alpha * alpha;
    float sin_v, cos_v;

    sin_v = sin(angle);
    cos_v = cos(angle);

    return vec2(cos_v, sin_v) * ssR;
}
#ifdef PBR_REFLECTIONS

    #include "/lib/reflection_lib.glsl"
#endif
vec3 Reproject(vec3 position)
{
    bool isHand = position.z < 0.6;
    if (position.z < 1.0 && !isHand)
    {
        position = ViewSpaceFromScreenSpace(position, gbufferProjectionInverse);
        position = mat3(gbufferModelViewInverse) * position + gbufferModelViewInverse[3].xyz;
        position += (cameraPosition - previousCameraPosition);
        position = mat3(gbufferPreviousModelView) * position + gbufferPreviousModelView[3].xyz;
        position = ScreenSpaceFromViewSpace(position, gbufferPreviousProjection);
    }
    else
    {
        position = ViewSpaceFromScreenSpace(position, gbufferProjectionInverse);
        position = mat3(gbufferModelViewInverse) * position + gbufferModelViewInverse[3].xyz;
        position = mat3(gbufferPreviousModelView) * position + gbufferPreviousModelView[3].xyz;
        position = ScreenSpaceFromViewSpace(position, gbufferPreviousProjection);
    }

    return position;
}
float MinDepth3x3(vec2 uv)
{
    ivec2 fragCoord = ivec2(gl_FragCoord.xy);
    float depth = texelFetch(depthtex0, fragCoord, 0).x;

    for (int x = -1; x <= 1; ++x)
    {
        for (int y = -1; y <= 1; ++y)
        {
            if (x != 0 || y != 0)
            {
                depth = min(depth, texelFetch(depthtex0, fragCoord + ivec2(x, y), 0).x);
            }
        }
    }

    return depth;
}


void main()
{
    /* RENDERTARGETS: 12 */

    vec2 texSize = viewResolution / textureSize(colortex12, 0); // Get the size of the base mipmap level (LOD 0)

    float z = texture(depthtex0, texcoord).x;
    bool hand = texture(depthtex2, texcoord).x > z;
#ifdef TAA
    vec2 tempOffset = offsets[framemod8] * texelSize * 0.5;
    vec3 fragpos = toScreenSpace(vec3(texcoord - tempOffset, z));
#else
    vec3 fragpos = toScreenSpace(vec3(texcoord, z));
#endif




    vec4 materials = texelFetch(colortex3, ivec2(gl_FragCoord.xy * texSize), 0);
    vec4 normalTex = texelFetch(colortex2, ivec2(gl_FragCoord.xy * texSize), 0);

    vec2 decoded = decodeVec2(normalTex.a);
    vec2 decoded2 = decodeVec2(materials.x);
    float reflectivity = decoded2.x;
    float smoothness = decoded2.y;
    vec3 scene = texture(colortex0, texcoord).rgb;
    if (z >= 1 || smoothness < REFLECTION_CUTOFF)
    {
        gl_FragData[0].rgb = vec3(0.0);
        return;
    }
    vec4 reflections = vec4(0);

    vec3 normal = (decodeNormal(normalTex.rg));
    vec2 lightmap = decodeVec2(normalTex.b);

    vec3 f0 = vec3(reflectivity);
    reflections = computeReflections(lightmap.y, smoothness, f0, normal, fragpos, decoded.y, scene, hand);
    //  reflections = clamp(reflections, 0.0, 512.0);
    int metalidx = int((f0.y * 255));
    gl_FragData[0] = reflections;
}
