

varying vec2 texcoord;

#define MIN(a, b) (dot(a, vec3(0.299, 0.587, 0.114)) < dot(b, vec3(0.299, 0.587, 0.114)) ? a : b)
#define MAX(a, b) (dot(a, vec3(0.299, 0.587, 0.114)) > dot(b, vec3(0.299, 0.587, 0.114)) ? a : b)
#define MED3(a, b, c) MAX(MIN(a, b), MIN(MAX(a, b), c))

#define SIGMA 10.0
#define BSIGMA 0.1
#define MSIZE 15
#define NSIGMA 0.01 // Normal similarity factor
#define DSIGMA 0.01 // Depth similarity factor

float normpdf(in float x, in float sigma)
{
    return 0.39894 * exp(-0.5 * x * x / (sigma * sigma)) / sigma;
}

float normpdf3(in vec3 v, in float sigma)
{
    return 0.39894 * exp(-0.5 * dot(v, v) / (sigma * sigma)) / sigma;
}

// Bilateral filter function that takes a sampler as input and is aware of normals and depth
vec3 bilateralFilter(in vec2 fragCoord, in vec2 resolution, sampler2D colorSampler)
{
    // Fetch the color, normal, and depth of the current fragment
    vec3 c = texture(colorSampler, (fragCoord.xy / resolution)).rgb;
    vec3 n = decodeNormal(texelFetch(colortex2, ivec2(fragCoord.xy), 0).rg);

    float d = texture(depthtex0, (fragCoord.xy / resolution)).r;

    // Declare variables for filter computation
    const int kSize = (MSIZE - 1) / 2;
    float kernel[MSIZE];
    vec3 final_colour = vec3(0.0);

    // Create the 1-D Gaussian kernel
    float Z = 0.0;
    for (int j = 0; j <= kSize; ++j)
    {
        kernel[kSize + j] = kernel[kSize - j] = normpdf(float(j), SIGMA);
    }

    // Normalize factor for the range kernel
    float bZ = 1.0 / normpdf(0.0, BSIGMA);

    // Apply the bilateral filter with normal and depth awareness
    for (int i = -kSize; i <= kSize; ++i)
    {
        for (int j = -kSize; j <= kSize; ++j)
        {
            // Offset coordinates
            vec2 offset = vec2(float(i), float(j));

            // Fetch neighboring color, normal, and depth
            vec3 cc = texture(colorSampler, (fragCoord.xy + offset) / resolution).rgb;
            // vec3 nc = texture(normalSampler, (fragCoord.xy + offset)).rgb;
            vec3 nc = decodeNormal(texelFetch(colortex2, ivec2((fragCoord.xy + offset)), 0).rg);

            float dc = texture(depthtex0, (fragCoord.xy + offset) / resolution).r;

            // Compute similarity weights
            float colorWeight = normpdf3(cc - c, BSIGMA) * bZ;
            float normalWeight = normpdf3(nc - n, NSIGMA);
            float depthWeight = normpdf(dc - d, DSIGMA);

            // Total weight combining color, normal, and depth similarities
            float weight = colorWeight * normalWeight * depthWeight * kernel[kSize + j] * kernel[kSize + i];
            Z += weight;
            final_colour += weight * cc;
        }
    }

    // Return the filtered color
    return final_colour / Z;
}

vec3 median(sampler2D tex1, ivec2 coords)
{
    vec3 v[9];

    // Fill the array 'v' with the color values from the 3x3 pixel window
    for (int dX = -1; dX <= 1; ++dX)
    {
        for (int dY = -1; dY <= 1; ++dY)
        {
            ivec2 offset = ivec2(dX, dY);
            v[(dX + 1) * 3 + (dY + 1)] = texelFetch(tex1, coords + offset, 0).rgb;
        }
    }

    // Calculate the median using the median of medians technique
    vec3 med1 = MED3(v[0], v[1], v[2]);
    vec3 med2 = MED3(v[3], v[4], v[5]);
    vec3 med3 = MED3(v[6], v[7], v[8]);
    vec3 median = MED3(med1, med2, med3);

    return median; // Median value (5th smallest)
}

vec3 tonemap(vec3 col)
{
    return col / (1 + luma(col));
}
vec3 invTonemap(vec3 col)
{
    return col / (1 - luma(col));
}
vec3 rgb2xyz(vec3 rgb)
{
    float r = rgb.r / 255.0;
    float g = rgb.g / 255.0;
    float b = rgb.b / 255.0;

    r = (r > 0.04045) ? pow((r + 0.055) / 1.055, 2.4) : r / 12.92;
    g = (g > 0.04045) ? pow((g + 0.055) / 1.055, 2.4) : g / 12.92;
    b = (b > 0.04045) ? pow((b + 0.055) / 1.055, 2.4) : b / 12.92;

    r *= 100.0;
    g *= 100.0;
    b *= 100.0;

    float x = r * 0.4124 + g * 0.3576 + b * 0.1805;
    float y = r * 0.2126 + g * 0.7152 + b * 0.0722;
    float z = r * 0.0193 + g * 0.1192 + b * 0.9505;

    return vec3(x, y, z);
}

vec3 xyz2lab(vec3 xyz)
{
    float x = xyz.x / 95.047;
    float y = xyz.y / 100.0;
    float z = xyz.z / 108.883;

    x = (x > 0.008856) ? pow(x, 1.0 / 3.0) : (7.787 * x) + (16.0 / 116.0);
    y = (y > 0.008856) ? pow(y, 1.0 / 3.0) : (7.787 * y) + (16.0 / 116.0);
    z = (z > 0.008856) ? pow(z, 1.0 / 3.0) : (7.787 * z) + (16.0 / 116.0);

    float l = (116.0 * y) - 16.0;
    float a = 500.0 * (x - y);
    float b = 200.0 * (y - z);

    return vec3(l, a, b);
}

vec3 rgb2lab(vec3 rgb)
{
    return xyz2lab(rgb2xyz(rgb));
}
vec3 lab2xyz(vec3 lab)
{
    float y = (lab.x + 16.0) / 116.0;
    float x = lab.y / 500.0 + y;
    float z = y - lab.z / 200.0;

    x = (pow(x, 3.0) > 0.008856) ? pow(x, 3.0) : (x - 16.0 / 116.0) / 7.787;
    y = (pow(y, 3.0) > 0.008856) ? pow(y, 3.0) : (y - 16.0 / 116.0) / 7.787;
    z = (pow(z, 3.0) > 0.008856) ? pow(z, 3.0) : (z - 16.0 / 116.0) / 7.787;

    x *= 95.047;
    y *= 100.0;
    z *= 108.883;

    return vec3(x, y, z);
}

vec3 xyz2rgb(vec3 xyz)
{
    float x = xyz.x / 100.0;
    float y = xyz.y / 100.0;
    float z = xyz.z / 100.0;

    float r = x * 3.2406 + y * -1.5372 + z * -0.4986;
    float g = x * -0.9689 + y * 1.8758 + z * 0.0415;
    float b = x * 0.0557 + y * -0.2040 + z * 1.0570;

    r = (r > 0.0031308) ? 1.055 * pow(r, 1.0 / 2.4) - 0.055 : 12.92 * r;
    g = (g > 0.0031308) ? 1.055 * pow(g, 1.0 / 2.4) - 0.055 : 12.92 * g;
    b = (b > 0.0031308) ? 1.055 * pow(b, 1.0 / 2.4) - 0.055 : 12.92 * b;

    return vec3(r, g, b) * 255.0;
}

vec3 lab2rgb(vec3 lab)
{
    return xyz2rgb(lab2xyz(lab));
}

float MinDepth3x3(vec2 uv)
{
    ivec2 fragCoord = ivec2(gl_FragCoord.xy);
    float depth = texelFetch(depthtex0, fragCoord, 0).x;

    for (int x = -1; x <= 1; ++x)
    {
        for (int y = -1; y <= 1; ++y)
        {
            if (x != 0 || y != 0)
            {
                depth = min(depth, texelFetch(depthtex0, fragCoord + ivec2(x, y), 0).x);
            }
        }
    }

    return depth;
}
vec3 untonemap(vec3 inp)
{
    return invTonemap(inp);
}
vec4 textureYC(sampler2D tex, vec2 uv)
{
    vec4 c = texture2D(tex, uv);
    //  c.rgb = median(tex, ivec2(uv * viewResolution));

    return vec4(tonemap(c.rgb), c.a);
}
vec4 texelFetchYC(sampler2D tex, ivec2 uv, float numb)
{
    vec4 c = texelFetch(tex, ivec2(uv), 0);
    //    c.rgb = median(tex, uv);

    return vec4(tonemap(c.rgb), c.a);
}
#ifdef DISTANT_HORIZONS

float MinDepth3x3DH(vec2 uv)
{
    ivec2 fragCoord = ivec2(gl_FragCoord.xy);
    float depth = texelFetch(dhDepthTex0, fragCoord, 0).x;

    for (int x = -1; x <= 1; ++x)
    {
        for (int y = -1; y <= 1; ++y)
        {
            if (x != 0 || y != 0)
            {
                depth = min(depth, texelFetch(dhDepthTex0, fragCoord + ivec2(x, y), 0).x);
            }
        }
    }

    return depth;
}

vec3 ReprojectDH(vec3 position)
{
    bool isHand = position.z < 0.6;
    if (position.z < 1.0 && !isHand)
    {
        position = ViewSpaceFromScreenSpace(position, dhProjectionInverse);
        position = mat3(gbufferModelViewInverse) * position + gbufferModelViewInverse[3].xyz;
        position += (cameraPosition - previousCameraPosition);
        position = mat3(gbufferPreviousModelView) * position + gbufferPreviousModelView[3].xyz;
        position = ScreenSpaceFromViewSpace(position, dhPreviousProjection);
    }
    else
    {
        position = ViewSpaceFromScreenSpace(position, dhProjectionInverse);
        position = mat3(gbufferModelViewInverse) * position + gbufferModelViewInverse[3].xyz;
        position = mat3(gbufferPreviousModelView) * position + gbufferPreviousModelView[3].xyz;
        position = ScreenSpaceFromViewSpace(position, dhPreviousProjection);
    }

    return position;
}
#endif
vec3 Reproject(vec3 position)
{
    bool isHand = position.z < 0.6;
    if (position.z < 1.0 && !isHand)
    {
        position = ViewSpaceFromScreenSpace(position, gbufferProjectionInverse);
        position = mat3(gbufferModelViewInverse) * position + gbufferModelViewInverse[3].xyz;
        position += (cameraPosition - previousCameraPosition);
        position = mat3(gbufferPreviousModelView) * position + gbufferPreviousModelView[3].xyz;
        position = ScreenSpaceFromViewSpace(position, gbufferPreviousProjection);
    }
    else
    {
        position = ViewSpaceFromScreenSpace(position, gbufferProjectionInverse);
        position = mat3(gbufferModelViewInverse) * position + gbufferModelViewInverse[3].xyz;
        position = mat3(gbufferPreviousModelView) * position + gbufferPreviousModelView[3].xyz;
        position = ScreenSpaceFromViewSpace(position, gbufferPreviousProjection);
    }

    return position;
}

vec3 clip_aabb(vec3 min_color, vec3 max_color, vec3 history_color)
{
    vec3 p_clip = 0.5 * (max_color + min_color);
    vec3 e_clip = 0.5 * (max_color - min_color);

    vec3 v_clip = history_color - p_clip;
    vec3 v_unit = v_clip / e_clip;
    vec3 a_unit = abs(v_unit);
    float ma_unit = max(a_unit.x, max(a_unit.y, a_unit.z));
    bool history_clipped = ma_unit > 1.0;
    vec3 history = history_clipped ? p_clip + v_clip / ma_unit : history_color;

    return history;
}
vec3 clip_aabb_simple(vec3 min_color, vec3 max_color, vec3 history_color)
{
    float tolerance = 0.0;

    // Expand the AABB by the tolerance value
    vec3 min_color_tolerant = min_color - tolerance;
    vec3 max_color_tolerant = max_color + tolerance;

    // Clamp the history color to the extended bounds of the AABB
    vec3 clipped_color = clamp(history_color, min_color_tolerant, max_color_tolerant);

    // Return the clipped color
    return clipped_color;
}

vec3 sampleBlur(
    sampler2D tex,
    vec2 res,
    vec2 pos,
    float sceneDepth,
    vec3 sceneNormal,
    float directions,
    float quality,
    float size)
{
    const float TAU = PI * 2.0;

    vec2 radius = size / res.xy;
    vec2 uv = pos / res.xy;

    vec3 color = texture(tex, uv).rgb;

    float directionStep = TAU / directions;
    float qualityStep = 1.0 / quality;

    float importance = 1.0;

    for (float arc = 0.0; arc < TAU; arc += directionStep)
        for (float q = 1.0; q > 0.0; q -= qualityStep)
        {
            vec2 offset = vec2(cos(arc), sin(arc)) * q * radius * res.y;
            vec2 cuv = uv + offset;

            // Avoid sampling outside the edges
            if (
                cuv.x < 0.0 ||
                cuv.y < 0.0 ||
                cuv.x > 1.0 ||
                cuv.y > 1.0)
            {
                continue;
            }

            vec4 fog = texture(tex, cuv);
            vec3 sampleNormal = decodeNormal(texture(colortex2, cuv).rg);
            float spatialImportance = 1.0 + (1.0 - q); // linear falloff
            float depthImportance = clamp(1.0 / abs(sceneDepth - fog.w), 0.0, 1.0);

            float normalDot = max(dot(sceneNormal, sampleNormal), 0.0);
            float normalImportance = normalDot; // using the dot product for normal importance

            float sampleImportance = spatialImportance * depthImportance * normalImportance;
            importance += sampleImportance;
            color += texture(tex, cuv).rgb * sampleImportance;
        }

    return color / importance;
}
void main()
{

    /* DRAWBUFFERS:1*/
    gl_FragData[0].rgb = tonemap(bilateralFilter(gl_FragCoord.xy, viewResolution, colortex1));


}
