#extension GL_ARB_gpu_shader5 : enable
#extension GL_ARB_shader_image_load_store : enable
#extension GL_ARB_shading_language_packing : enable

flat varying vec3 ambientUp;
flat varying vec3 ambientDown;
flat varying vec3 ambientLeft;
flat varying vec3 ambientRight;
flat varying vec3 ambientB;
flat varying vec3 ambientF;
flat varying float ambientA;
varying vec2 texcoord;

#include "/lib/light.glsl"

#ifdef LPV_ENABLED
    #include "/lib/hsv.glsl"
    #include "/lib/lpv/lpv_common.glsl"
    #include "/lib/lpv/lpv_render.glsl"
#endif

float linZ(float depth)
{
    return (2.0 * near) / (far + near - depth * (far - near));
}

#if SHADOW_FILTERING == 2
    #include "/lib/shadows/shadowfunc_advanced.glsl"

#else
    #include "/lib/shadows/shadowfunc_simple.glsl"
#endif

vec3 GGX2(vec3 normal, vec3 viewDir, vec3 lightDir, float roughness, vec3 baseReflectance)
{
    // Calculate the half-vector (H), which is halfway between the view direction and light direction
    vec3 halfVector = normalize(lightDir + viewDir);

    // Dot products for various terms
    float NdotH = max(dot(normal, halfVector), 0.0);  // Dot product between normal and half-vector
    float NdotL = max(dot(normal, lightDir), 0.0);    // Dot product between normal and light direction
    float NdotV = max(dot(normal, viewDir), 0.0);     // Dot product between normal and view direction
    float VdotH = max(dot(viewDir, halfVector), 0.0); // Dot product between view direction and half-vector

    // ----------------------------
    // Fresnel term (F)
    // ----------------------------
    // Schlick's approximation for the Fresnel effect
    float oneMinusVdotH = 1.0 - VdotH;
    float fresnelFactor = oneMinusVdotH * oneMinusVdotH * oneMinusVdotH * oneMinusVdotH * oneMinusVdotH; // (1 - VdotH)^5
    vec3 fresnel = mix(baseReflectance, vec3(1.0), fresnelFactor);                                       // F = F0 + (1 - F0) * (1 - VdotH)^5

    // ----------------------------
    // Distribution term (D)
    // ----------------------------
    // GGX normal distribution function (NDF)
    float adjustedRoughness = roughness + 1e-4;          // Ensure roughness is never zero to avoid division by zero
    float alpha = adjustedRoughness * adjustedRoughness; // Squared roughness
    float NdotH2 = NdotH * NdotH;                        // Square of the dot product between normal and half-vector
    float distributionDenom = NdotH2 * (alpha - 1.0) + 1.0;
    float distributionTerm = alpha / (distributionDenom * distributionDenom * PI);

    // ----------------------------
    // Geometry term (G)
    // ----------------------------
    // Smith's geometry function with Schlick-GGX approximation
    float k = (adjustedRoughness + 1.0) * (adjustedRoughness + 1.0) * 0.125; // k^2 / 8
    float geometryFactor = 1.0 - k;

    float geomVisibilityV = NdotV / (NdotV * geometryFactor + k); // Geometry term for view direction
    float geomVisibilityL = NdotL / (NdotL * geometryFactor + k); // Geometry term for light direction

    // ----------------------------
    // Final Specular Calculation
    // ----------------------------
    // Combine all terms: Distribution (D), Geometry (G), and Fresnel (F)
    float denominator = max(4.0 * NdotV * NdotL, 1e-7); // Prevent division by zero with a small epsilon
    vec3 specular = (distributionTerm * geomVisibilityV * geomVisibilityL * fresnel) * (NdotL / denominator);

    return specular; // Return the computed specular reflection
}

vec2 tapLocation(int sampleNumber, float spinAngle, int nb, float nbRot)
{
    float alpha = float(sampleNumber + spinAngle / 6.28) / nb;
    float angle = alpha * (nbRot * 6.28) + spinAngle;

    float ssR = alpha * alpha;
    float sin_v, cos_v;

    sin_v = sin(angle);
    cos_v = cos(angle);

    return vec2(cos_v, sin_v) * ssR;
}
#if SSAO == 2
    #include "/lib/rtao_lib.glsl"
#elif SSAO == 1
    #include "/lib/rtao_lib_simple.glsl"
#endif

float getRawDepth(vec2 uv)
{
    return texture(depthtex0, uv).x;
}

vec3 viewSpacePosAtScreenUV(vec2 uv)
{
    vec3 viewSpaceRay = (gbufferProjectionInverse * vec4(uv * 2.0 - 1.0, 1.0, 1.0) * near).xyz;
    float rawDepth = getRawDepth(uv);
    return viewSpaceRay * (linZ(rawDepth));
}

vec3 viewNormalAtPixelPosition2(vec2 vpos)
{
    // screen uv from vpos
    vec2 uv = vpos * texelSize.xy;

    // current pixel's depth
    float c = getRawDepth(uv);

    // get current pixel's view space position
    vec3 viewSpacePos_c = viewSpacePosAtScreenUV(uv);

    // get view space position at 1 pixel offsets in each major direction
    vec3 viewSpacePos_l = viewSpacePosAtScreenUV(uv + vec2(-1.0, 0.0) * texelSize.xy);
    vec3 viewSpacePos_r = viewSpacePosAtScreenUV(uv + vec2(1.0, 0.0) * texelSize.xy);
    vec3 viewSpacePos_d = viewSpacePosAtScreenUV(uv + vec2(0.0, -1.0) * texelSize.xy);
    vec3 viewSpacePos_u = viewSpacePosAtScreenUV(uv + vec2(0.0, 1.0) * texelSize.xy);

    // get the difference between the current and each offset position
    vec3 l = viewSpacePos_c - viewSpacePos_l;
    vec3 r = viewSpacePos_r - viewSpacePos_c;
    vec3 d = viewSpacePos_c - viewSpacePos_d;
    vec3 u = viewSpacePos_u - viewSpacePos_c;

    // get depth values at 1 & 2 pixels offsets from current along the horizontal
    // axis
    vec4 H = vec4(getRawDepth(uv + vec2(-1.0, 0.0) * texelSize.xy), getRawDepth(uv + vec2(1.0, 0.0) * texelSize.xy),
                  getRawDepth(uv + vec2(-2.0, 0.0) * texelSize.xy), getRawDepth(uv + vec2(2.0, 0.0) * texelSize.xy));

    // get depth values at 1 & 2 pixels offsets from current along the vertical
    // axis
    vec4 V = vec4(getRawDepth(uv + vec2(0.0, -1.0) * texelSize.xy), getRawDepth(uv + vec2(0.0, 1.0) * texelSize.xy),
                  getRawDepth(uv + vec2(0.0, -2.0) * texelSize.xy), getRawDepth(uv + vec2(0.0, 2.0) * texelSize.xy));

    // current pixel's depth difference from slope of offset depth samples
    // differs from original article because we're using non-linear depth values
    // see article's comments
    vec2 he = abs((2 * H.xy - H.zw) - c);
    vec2 ve = abs((2 * V.xy - V.zw) - c);

    // pick horizontal and vertical diff with the smallest depth difference from
    // slopes
    vec3 hDeriv = he.x < he.y ? l : r;
    vec3 vDeriv = ve.x < ve.y ? d : u;

    // get view space normal from the cross product of the best derivatives
    vec3 viewNormal = normalize(cross(hDeriv, vDeriv));

    return viewNormal;
}
// Function to calculate the torch lighting with optional LPV support
vec3 CalculateTorch(
    vec2 lightmap,
    vec3 normal,
    vec3 fragpos,
    float ao)
{
    // Initialize torch color
    vec3 torch = vec3(3.000, 1.382, 0.408);

    // Apply lightmap power
    float lightmapPow = pow(lightmap.x, 4.0);
    torch *= lightmapPow + lightmap.x * 0.025;
    torch = clamp(torch, 0.0, 4.0) * 2.0;

#ifdef LPV_ENABLED
    // Calculate normal offset for LPV
    vec3 normalOffset = normal * (LPV_NORMAL_STRENGTH * 0.01);

    // Get LPV position
    vec3 lpvPos = GetLpvPosition(mat3(gbufferModelViewInverse) * fragpos + gbufferModelViewInverse[3].xyz) + normalOffset;

    // Calculate LPV torch light
    vec3 LpvTorchLight = GetLpvBlockLight(SampleLpvLinear(lpvPos)) * (1.0 + (1.0 - ao));
    LpvTorchLight *= clamp(lightmap.x * lightmap.x + lightmap.x * 0.025, 0.0, 10.0) * 1.5;

    // Calculate fade factor based on distance
    vec3 cubicRadius = clamp(min(((LpvSize3 - 1.0) - lpvPos) / LPV_BLEND_DISTANCE, lpvPos / LPV_BLEND_DISTANCE), 0.0, 1.0);
    float LpvFadeF = cubicRadius.x * cubicRadius.y * cubicRadius.z;

    // Smooth the fade factor
    LpvFadeF = 1.0 - pow(1.0 - pow(LpvFadeF, 1.5), 3.0); // make it nice and soft :)
    #ifdef DYNAMIC_LIGHTS_COMP
    LpvFadeF = clamp(1 - pow(distance(LpvTorchLight, torch), 0.25), 0, 1);
    #endif
    // Blend torch with LPV torch light based on fade factor
    torch = mix(torch, LpvTorchLight, LpvFadeF);
#endif

    return torch;
}

void decodeBasicMaterial(sampler2D inputtex, ivec2 texCoords,
                         out vec3 normal, out float ao, out float sss, out float porosity, out vec2 lightmap)
{
    vec4 normalTex = texelFetch(inputtex, texCoords, 0);
    normal = decodeNormal(normalTex.rg);
    vec2 decoded = decodeVec2(normalTex.a);
    float threshold = 64.5 / 255.0;
    sss = (decoded.y > threshold) ? (decoded.y - threshold) / (1.0 - threshold) : 0.0;
    sss *= 0.75;
    porosity = (decoded.y < threshold) ? (decoded.y / threshold) : 0.0;

    lightmap = decodeVec2(normalTex.b);
    lightmap.x = pow(lightmap.x, 3.0);

    ao = decoded.x;
#ifdef LPV_ENABLED
    ao = mix(ao, 1.0, 0.5);
#endif
}

void decodeExtendedMaterial(sampler2D inputtex1, ivec2 texCoords, sampler2D inputtex2, vec2 texcoord, float wetness, vec3 albedo,
                            out vec3 normal, out float ao, out float sss, out vec2 lightmap, out vec3 reflectivity, out float roughness)
{
    float porosity;
    decodeBasicMaterial(inputtex1, texCoords, normal, ao, sss, porosity, lightmap);

#ifdef PBR
    vec4 materials = texture(inputtex2, texcoord);
    vec2 decoded2 = decodeVec2(materials.x);
    float rainMult = sqrt(lightmap.y) * wetness * (1.0 - (porosity * porosity));
    roughness = mix(pow(1.0 - decoded2.y, 2.0), 0.01, rainMult);
    reflectivity = vec3(decoded2.x);
    reflectivity = (int(reflectivity.x * 255.0) > 229) ? albedo : reflectivity;

#endif
}
float gbf(in vec3 eyePlayerPos, in float far, in float z0)
{
    float distXZ = length((vec3(eyePlayerPos.x, 0.0, eyePlayerPos.z)).xyz);
    float distY = length((vec3(0.0, eyePlayerPos.y, 0.0)).xyz);
    float eyeDist = max(distXZ, distY);
#ifdef DISTANT_HORIZONS
    float borderFogFactor = smoothstep(0.0, dhRenderDistance, eyeDist);
    borderFogFactor *= borderFogFactor * borderFogFactor;
    return borderFogFactor;
#else

    float borderFogFactor = smoothstep(far * 0.5, far, eyeDist);
    borderFogFactor *= borderFogFactor * borderFogFactor;

    return borderFogFactor;

#endif
}

vec3 CalculateWaterCaustics(vec3 screenSpacePosition)
{

    vec3 worldPos = mat3(gbufferModelViewInverse) * screenSpacePosition + cameraPosition;
    vec2 pos = (worldPos.xz - lightPos.xz / lightPos.y * worldPos.y);

    vec2 coord0 = mod((pos) / 128.0, 1.0);
    vec2 coord1 = mod((pos) / 16.0, 1.0);

    vec3 value = vec3(0);
    value.rgb += texture(colortex7, coord0).g;

    value.rgb += texture(colortex7, coord1).g;

    return value * 4.0 + 0.5;
}
float ld(const float dist)
{
    return (2.0 * near) / (far + near - dist * (far - near));
}

void main()
{

    vec3 lightPosition = lightPos;

#ifndef OVERWORLD
    lightPosition = vec3(0.0, 1.0, 0.0);
#endif
#ifdef PBR_REFLECTIONS
/* DRAWBUFFERS:01 */
#else
/* DRAWBUFFERS:0 */
#endif

    // Texture Coordinates
    ivec2 texCoords = ivec2(gl_FragCoord.xy);

    // Depth Check
    float z = texture(depthtex0, texcoord).x;
    if (z >= 1.0)
    {

        gl_FragData[0].rgb = texture(colortex8, texcoord).rgb;
#ifndef IS_IRIS
        if (!(texelFetch(depthtex2, texCoords, 0).x > z))
            gl_FragData[0].rgb += texture(colortex2, texcoord).rgb;
#endif
        return;
    }

    // Material Properties
    vec3 albedo = texelFetch(colortex0, texCoords, 0).rgb;
    bool hand = texelFetch(depthtex2, texCoords, 0).x > z;

    // Declare material variables
    vec3 normal, reflectivity;
    float ao, sss, roughness;
    vec2 lightmap;

    // Decode material properties
    decodeExtendedMaterial(colortex2, texCoords, colortex3, texcoord, wetness, albedo, normal, ao, sss, lightmap, reflectivity, roughness);

    // Compute fragpos
    vec3 fragpos;
#ifdef TAA
    fragpos = toScreenSpace(vec3(texcoord - (offsets[framemod8] * texelSize * 0.5), z));
#else
    fragpos = toScreenSpace(vec3(texcoord, z));
#endif

    // Compute p3
    vec3 p3 = mat3(gbufferModelViewInverse) * fragpos + gbufferModelViewInverse[3].xyz;

    // Initialize variables
    vec3 colored = vec3(0.0);
    vec3 sAlbedo = vec3(0.6) * ambientA * lightmap.y;
    float entity = texture(colortex5, texcoord).g;
    float diffuseSun = max(dot(normal, lightPosition), 0.0);
    vec3 specular = vec3(0.0);

    // Lighting Calculations
    if (!hand)
    {
// Ambient Occlusion
#if SSAO > 0
        ao *= rtao(lightmap.y, normal, fragpos);
#else
        ao = pow(ao, 2.0);
#endif

// Shadow Filtering
#ifdef OVERWORLD
    #if SHADOW_FILTERING == 2
        diffuseSun = shadowFunc(diffuseSun, sss, p3, noise_standard(gl_FragCoord.xy), lightmap, fragpos, sAlbedo, albedo, normal, colored, entity);
    #else
        diffuseSun = shadowFunc(diffuseSun, sss, p3, noise_standard(gl_FragCoord.xy), lightmap, normal);
    #endif
#endif

// Specular Calculation
#ifndef PBR_REFLECTIONS
    #ifdef PBR
        specular = GGX2(normal, normalize(-p3), lightPosition, roughness, reflectivity) * diffuseSun * sunLight;
    #endif
#endif
    }
    else
    {
        diffuseSun = max(dot(viewToWorld(normal), lightPosition), 0.0);
        diffuseSun *= min(eyeBrightnessSmooth.y * recip * lightmap.y, 1.0);
    }

    // Torch Lighting and Ambient Light
    vec3 torchLighting = CalculateTorch(lightmap, normal, fragpos, ao);
    vec3 ambient = calcAmbientLight(normal, lightmap.xy, ambientUp, ambientDown, ambientRight, ambientLeft, ambientB, ambientF, sAlbedo + colored, diffuseSun, torchLighting) * ao;

// Caustics Calculation
#ifdef CAUSTICS
    if (!hand)
    {
        vec3 fragposition = mat3(gbufferModelViewInverse) * fragpos + gbufferModelViewInverse[3].xyz + (cameraPosition - previousCameraPosition);
        vec3 previousPosition = mat3(gbufferPreviousModelView) * fragposition + gbufferPreviousModelView[3].xyz;
        previousPosition = toClipSpace3Prev(previousPosition);

        vec2 velocity = previousPosition.xy - texcoord;
        previousPosition.xy = texcoord + velocity;

        float prevmask = texture(colortex4, previousPosition.xy).a;

        bool condition = (isEyeInWater == 0 && prevmask > 0.0) || (isEyeInWater == 1 && prevmask <= 0.0);
        if (condition)
        {
            vec3 caustics = CalculateWaterCaustics(fragpos);
            ambient *= mix(vec3(1.0), caustics, lightmap.y);
        }
    }
#endif

// Final Color Composition
#ifdef PBR_REFLECTIONS
    #ifdef HARDCODED_METALS
    vec3 color = ambient * albedo;
    #else
    vec3 color = ambient * albedo + specular;
    #endif
#else
    vec3 color = ambient * albedo + specular;
#endif

// Debugging Outputs
#if DEBUG_MODE == 1
    color = ambient;
#elif DEBUG_MODE == 2
#elif DEBUG_MODE == 3
    color = normal;
#elif DEBUG_MODE == 4
    color = vec3(diffuseSun);
#elif DEBUG_MODE == 5
    color = torch;
#endif

    // Fragment Outputs
    gl_FragData[0].rgb = color;

#ifdef PBR_REFLECTIONS
    gl_FragData[1].rgb = albedo;
    gl_FragData[1].a = diffuseSun;
#endif
}
