#extension GL_EXT_gpu_shader4 : enable
varying vec2 texcoord;

// YCoCg Color Space Conversion Functions
vec3 RGB_YCoCg(vec3 c)
{
#ifdef USE_YCOCG
    return vec3(
        c.r / 4.0 + c.g / 2.0 + c.b / 4.0,
        c.r / 2.0 - c.b / 2.0,
        -c.r / 4.0 + c.g / 2.0 - c.b / 4.0);
#else
    return c;
#endif
}

vec3 YCoCg_RGB(vec3 c)
{
#ifdef USE_YCOCG
    return vec3(
        c.x + c.y - c.z,
        c.x + c.z,
        c.x - c.y - c.z);
#else
    return c;
#endif
}

// Texture sampling functions with YCoCg conversion
vec4 textureYC(sampler2D tex, vec2 uv)
{
    vec4 c = texture2D(tex, uv);
    return vec4(RGB_YCoCg(c.rgb), c.a);
}

vec4 texelFetchYC(sampler2D tex, ivec2 uv, int level)
{
    vec4 c = texelFetch(tex, uv, level);
    return vec4(RGB_YCoCg(c.rgb), c.a);
}

// Fetch history pixel
vec4 fetchHistoryPx(ivec2 pix)
{
    return vec4(texelFetchYC(colortex4, pix, 0).xyz, 1.0);
}

// Cubic Hermite interpolation
vec3 CubicHermite(vec3 A, vec3 B, vec3 C, vec3 D, float t)
{
    float t2 = t * t;
    float t3 = t2 * t;
    vec3 a = -0.5 * A + 1.5 * B - 1.5 * C + 0.5 * D;
    vec3 b = A - 2.5 * B + 2.0 * C - 0.5 * D;
    vec3 c = -0.5 * A + 0.5 * C;
    vec3 d = B;
    return a * t3 + b * t2 + c * t + d;
}

// Bicubic Hermite texture sampling
vec3 BicubicHermiteTextureSample(vec2 P)
{
    vec2 pixel = P * viewResolution + 0.5;
    vec2 frac = fract(pixel);
    ivec2 ipixel = ivec2(pixel) - 1;

    vec3 C00 = fetchHistoryPx(ipixel + ivec2(-1, -1)).rgb;
    vec3 C10 = fetchHistoryPx(ipixel + ivec2(0, -1)).rgb;
    vec3 C20 = fetchHistoryPx(ipixel + ivec2(1, -1)).rgb;
    vec3 C30 = fetchHistoryPx(ipixel + ivec2(2, -1)).rgb;

    vec3 C01 = fetchHistoryPx(ipixel + ivec2(-1, 0)).rgb;
    vec3 C11 = fetchHistoryPx(ipixel + ivec2(0, 0)).rgb;
    vec3 C21 = fetchHistoryPx(ipixel + ivec2(1, 0)).rgb;
    vec3 C31 = fetchHistoryPx(ipixel + ivec2(2, 0)).rgb;

    vec3 C02 = fetchHistoryPx(ipixel + ivec2(-1, 1)).rgb;
    vec3 C12 = fetchHistoryPx(ipixel + ivec2(0, 1)).rgb;
    vec3 C22 = fetchHistoryPx(ipixel + ivec2(1, 1)).rgb;
    vec3 C32 = fetchHistoryPx(ipixel + ivec2(2, 1)).rgb;

    vec3 C03 = fetchHistoryPx(ipixel + ivec2(-1, 2)).rgb;
    vec3 C13 = fetchHistoryPx(ipixel + ivec2(0, 2)).rgb;
    vec3 C23 = fetchHistoryPx(ipixel + ivec2(1, 2)).rgb;
    vec3 C33 = fetchHistoryPx(ipixel + ivec2(2, 2)).rgb;

    vec3 CP0X = CubicHermite(C00, C10, C20, C30, frac.x);
    vec3 CP1X = CubicHermite(C01, C11, C21, C31, frac.x);
    vec3 CP2X = CubicHermite(C02, C12, C22, C32, frac.x);
    vec3 CP3X = CubicHermite(C03, C13, C23, C33, frac.x);

    return CubicHermite(CP0X, CP1X, CP2X, CP3X, frac.y);
}

// Compute minimum depth over a 3x3 neighborhood
float MinDepth3x3(sampler2D tex, ivec2 fragCoord)
{
    float depth = texelFetch(tex, fragCoord, 0).x;
    for (int x = -1; x <= 1; ++x)
    {
        for (int y = -1; y <= 1; ++y)
        {
            if (x != 0 || y != 0)
            {
                depth = min(depth, texelFetch(tex, fragCoord + ivec2(x, y), 0).x);
            }
        }
    }
    return depth;
}

#ifdef DISTANT_HORIZONS
float MinDepth3x3DH(sampler2D tex, ivec2 fragCoord)
{
    float depth = texelFetch(dhDepthTex0, fragCoord, 0).x;
    for (int x = -1; x <= 1; ++x)
    {
        for (int y = -1; y <= 1; ++y)
        {
            if (x != 0 || y != 0)
            {
                depth = min(depth, texelFetch(dhDepthTex0, fragCoord + ivec2(x, y), 0).x);
            }
        }
    }
    return depth;
}

vec3 ReprojectDH(vec3 position)
{
    bool isHand = position.z < 0.6;
    if (position.z < 1.0 && !isHand)
    {
        position = ViewSpaceFromScreenSpace(position, dhProjectionInverse);
        position = mat3(gbufferModelViewInverse) * position + gbufferModelViewInverse[3].xyz;
        position += (cameraPosition - previousCameraPosition);
        position = mat3(gbufferPreviousModelView) * position + gbufferPreviousModelView[3].xyz;
        position = ScreenSpaceFromViewSpace(position, dhPreviousProjection);
    }
    else
    {
        position = ViewSpaceFromScreenSpace(position, dhProjectionInverse);
        position = mat3(gbufferModelViewInverse) * position + gbufferModelViewInverse[3].xyz;
        position = mat3(gbufferPreviousModelView) * position + gbufferPreviousModelView[3].xyz;
        position = ScreenSpaceFromViewSpace(position, dhPreviousProjection);
    }
    return position;
}
#endif

vec3 Reproject(vec3 position)
{
    bool isHand = position.z < 0.6;

    position = ViewSpaceFromScreenSpace(position, gbufferProjectionInverse);

    if (position.z < 1.0 && !isHand)
    {
        position = mat3(gbufferModelViewInverse) * position + gbufferModelViewInverse[3].xyz;
        position += (cameraPosition - previousCameraPosition);
    }
    else
    {
        position = mat3(gbufferModelViewInverse) * position + gbufferModelViewInverse[3].xyz;
    }
    position = mat3(gbufferPreviousModelView) * position + gbufferPreviousModelView[3].xyz;

    position = ScreenSpaceFromViewSpace(position, gbufferPreviousProjection);

    return position;
}

// Simple clipping function for color history
vec3 clip_aabb_simple(vec3 min_color, vec3 max_color, vec3 history_color)
{
    return clamp(history_color, min_color, max_color);
}
/* DRAWBUFFERS:4 */

void main()
{
    vec3 color = vec3(0.0);
    vec2 uv = texcoord;
    ivec2 texel = ivec2(uv * viewResolution);

#ifdef TAA
    float z = texture(depthtex0, uv).x;
    float alpha = texture(colortex1, uv).a;
    float entity = texture(colortex5, uv).g;
    float offcenter_rej = TAA_OFFCENTER_REJECTION;
    float historyWeight = TAA_HISTORY_WEIGHT;

    float depth = MinDepth3x3(depthtex0, texel);
    vec3 position = vec3(uv, depth);

    float depth2 = MinDepth3x3(depthtex1, texel);
    vec3 position2 = vec3(uv, depth2);

    if (alpha > 0.1)
    {
        position = mix(position2, position, pow(alpha, 10.0));
    }

    vec3 reprojectedPosition = Reproject(position);

    #ifdef DISTANT_HORIZONS
    if (z >= 1.0)
    {
        depth = MinDepth3x3DH(dhDepthTex0, texel);
        position = vec3(uv, depth);
        reprojectedPosition = ReprojectDH(position);
        historyWeight = 0.98;
        offcenter_rej = 0.1;
    }
    #endif

    vec3 current = texture2D(colortex0, uv + offsets[framemod8] * texelSize * 0.5).rgb;
    vec3 history = BicubicHermiteTextureSample(reprojectedPosition.xy);

    if (any(lessThan(reprojectedPosition.xy, vec2(0.0))) || any(greaterThan(reprojectedPosition.xy, vec2(1.0))))
    {
        historyWeight = 0.0;
    }

    #ifdef TAA_CLIP
    vec2 pixelCenterDist = 1.0 - abs(2.0 * fract(reprojectedPosition.xy * viewResolution) - 1.0);
    historyWeight *= sqrt(pixelCenterDist.x * pixelCenterDist.y) * offcenter_rej + (1.0 - offcenter_rej);

    ivec2 du = ivec2(1, 0);
    ivec2 dv = ivec2(0, 1);

    // Fetch center pixel first
    vec3 cmc = texelFetch2D(colortex0, texel, 0).rgb;

    // Reuse values by minimizing fetches
    vec3 cml = texelFetch2D(colortex0, texel - du, 0).rgb;
    vec3 cmr = texelFetch2D(colortex0, texel + du, 0).rgb;
    vec3 ctc = texelFetch2D(colortex0, texel - dv, 0).rgb;
    vec3 cbc = texelFetch2D(colortex0, texel + dv, 0).rgb;

    // Fetch only the corners
    vec3 ctl = texelFetch2D(colortex0, texel - dv - du, 0).rgb;
    vec3 ctr = texelFetch2D(colortex0, texel - dv + du, 0).rgb;
    vec3 cbl = texelFetch2D(colortex0, texel + dv - du, 0).rgb;
    vec3 cbr = texelFetch2D(colortex0, texel + dv + du, 0).rgb;

    // Step 1: Compare the top row
    vec3 minTop = min(min(ctl, ctc), ctr);
    vec3 maxTop = max(max(ctl, ctc), ctr);

    // Step 2: Compare the middle row
    vec3 minMid = min(min(cml, cmc), cmr);
    vec3 maxMid = max(max(cml, cmc), cmr);

    // Step 3: Compare the bottom row
    vec3 minBot = min(min(cbl, cbc), cbr);
    vec3 maxBot = max(max(cbl, cbc), cbr);

    // Final reduction
    vec3 minRounded = min(minTop, min(minMid, minBot));
    vec3 maxRounded = max(maxTop, max(maxMid, maxBot));

    history = clip_aabb_simple(minRounded, maxRounded, history);
    #endif

    #ifdef NO_ENTITY_TAA
    if (entity > 0.1)
    {
        historyWeight = clamp(1.0 - entity, 0.15, historyWeight);
    }
    #endif

    color = mix(current, history, historyWeight);
#else
    color = texture(colortex0, uv).rgb;
#endif

    gl_FragData[0].rgb = clamp(color, 0.0001, 1000.0);
    gl_FragData[0].a = texture(colortex4, uv).a;
}
